(function (Drupal, drupalSettings) {
  /**
   * Process the activity response and updates the markup for each user.
   */
  function render_activity(response) {
    if (response.status === 'success') {
      let contributors_listed = document.querySelectorAll('.contributor-username');
      contributors_listed.forEach((contributor) => {
        const username = contributor.innerText;
        const activity = response.data.activity[username];
        if (activity !== undefined) {
          if (activity.comments === undefined) {
            activity.comments = 0;
          }
          if (activity.reactions === undefined) {
            activity.reactions = [];
          }
          if (activity.files === undefined) {
            activity.files = 0;
          }

          let user_wrapper = contributor.parentNode;
          let activity_text = [];
          if (activity.comments) {
            activity_text.push(Drupal.formatPlural(activity.comments, '1 comment', '@comments comments', {
              '@comments': activity.comments
            }));
          }
          if (activity.files) {
            activity_text.push(Drupal.formatPlural(activity.files, '1 file', '@files files', {
              '@files': activity.files
            }));
          }
          const reactions_length = activity.reactions.length;
          if (reactions_length) {
            let reactions_text = Drupal.formatPlural(reactions_length, 'Reaction given: ', 'Reactions given: ') + activity.reactions.slice(0, 5).join(' ');
            if (reactions_length > 5) {
              reactions_text += ' …';
            }
            activity_text.push(reactions_text);
          }

          if (activity_text.length) {
            activity_text = activity_text.join('<br>');
            const previous_activity = user_wrapper.querySelector('.issue_activity');
            if (previous_activity) {
              user_wrapper.removeChild(previous_activity);
            }
            user_wrapper.innerHTML = user_wrapper.innerHTML + '<div class="issue_activity description">' + activity_text + '</div>';
          }
        }
      });
    }
  }

  /**
   * Replace plain links with meta info like title, status, etc.
   */
  function add_meta_info_link(link_meta) {
    let link_element = document.querySelector('a.related-link[href="' + link_meta['link'] + '"]');
    if (link_element !== undefined) {
      link_element.setAttribute('target', '_blank');
      link_element.classList.add(link_meta['type']);
      link_element.textContent = link_meta['title'];
      if (link_meta['type'] === 'merge-request-gitlab') {
        const iid = link_meta['link'].split('/').pop();
        link_element.textContent = Drupal.t('Merge request #@iid: ', {'@iid': iid}) + link_element.textContent;
      }
      if (link_meta['status'] === 'closed') {
        link_element.insertAdjacentHTML('afterend', ' <small>' + Drupal.t('CLOSED') + '</small>');
      }
    }
  }

  /**
   * Sets a loading message and returns its id.
   */
  function set_loading_message(message) {
    const random_id = Math.floor(Math.random() * 100);
    const loading_message_wrapper_selector = drupalSettings.contribution_records.loading_wrapper_selector;
    let loading_message = '';
    if (loading_message_wrapper_selector) {
      loading_message = '<div id="loading_' + random_id + '" class="loading_activity">' + message + '</div>';
      document.querySelector(loading_message_wrapper_selector).innerHTML = loading_message + document.querySelector(loading_message_wrapper_selector).innerHTML;
    }

    return 'loading_' + random_id;
  }

  /**
   * Cleans a loading message.
   */
  function clean_loading_message(loading_id) {
    document.getElementById(loading_id).remove();
  }

  /**
   * Fetches activity from a URL and renders the results.
   */
  function fetch_activity(link) {
    // console.log(link);
    let url_params = new URLSearchParams();
    url_params.append('source_link', link);
    url_params.append('maintainer', drupalSettings.contribution_records.maintainer ? 1 : 0);
    return fetch(drupalSettings.path.baseUrl + 'contribution_record_source_activity?' + url_params)
      .then(response => response.json());
  }

  Drupal.behaviors.issue_activity = {
    attach: async function (context, settings) {
      const contributors_wrapper = once('contributors', 'body', context);
      if (contributors_wrapper.length) {
        let combined_data = {};
        let links_meta = [];
        const loading_id = set_loading_message(Drupal.t('Loading per-user activity information...'));
        fetch_activity(drupalSettings.contribution_records.source_link)
          .then(result => {
            if (result.status === 'success') {
              combined_data = result;
              render_activity(combined_data);
            }
            clean_loading_message(loading_id);
          })
          .then(() => {
            let related_links = document.querySelectorAll('.related-link');
            if (related_links.length) {
              const related_loading_id = set_loading_message(Drupal.t('Loading per-user activity information for related links'));
              Promise.all(Array.from(related_links).map(related_link =>
                fetch_activity(related_link.href)
              ))
              .then(results_array => {
                results_array.forEach(result => {
                  if (result.status === 'success') {
                    add_meta_info_link(result.data.meta);

                    // Merge / add data to the combined_data array.
                    Object.entries(result.data.activity).forEach(([key, value]) => {
                      if (combined_data.data.activity[key] === undefined) {
                        combined_data.data.activity[key] = value;
                      }
                      else {
                        combined_data.data.activity[key].comments += value.comments;
                        combined_data.data.activity[key].files += value.files;
                        if (value.reactions.length) {
                          combined_data.data.activity[key].reactions = combined_data.data.activity[key].reactions.concat(value.reactions);
                        }
                      }
                    });
                  }
                });
                render_activity(combined_data);
                clean_loading_message(related_loading_id);
              })
              .catch(error => {
                // console.log(error);
                clean_loading_message(related_loading_id);
              });
            }
          })
          .catch(error => {
            // console.log(error);
            clean_loading_message(loading_id);
          });
      }
    }
  }
} (Drupal, drupalSettings));
